@@ -1,4 +1,4 @@
-# Copyright 2015-2022 D.G. MacCarthy <https://dmaccarthy.github.io/sc8pr>
+# Copyright 2015-2023 D.G. MacCarthy <https://dmaccarthy.github.io/sc8pr>
 #
 # This file is part of "sc8pr".
 #
@@ -9,25 +9,25 @@
 #
 # "sc8pr" is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with "sc8pr".  If not, see <http://www.gnu.org/licenses/>.
+# along with "sc8pr". If not, see <http://www.gnu.org/licenses/>.
 
 
-version = 2, 2, 2
+version = 3, 0, 0
 print("sc8pr {}.{}.{}: https://dmaccarthy.github.io/sc8pr".format(*version))
 
-import sys, struct, zlib, io
+import sys, struct
 from math import hypot, sqrt
 import pygame
 import pygame.display as _pd
 from pygame.transform import flip as _pyflip
 from sc8pr._event import EventManager
 from sc8pr._cs import CoordSys
-from sc8pr.geom import transform2d, positiveAngle, delta, sigma, vmult
-from sc8pr.util import CachedSurface, style, logError, sc8prData, tile, rgba, drawBorder
+from sc8pr.geom import transform2d, delta, sigma, vmult, neg # , positiveAngle
+from sc8pr.util import CachedSurface, style, logError, sc8prData, resolvePath, tile, rgba, hasAlpha, surface, drawBorder, crop, export, customEv
 
 # Anchor point constants
 TOPLEFT = 0
@@ -50,105 +50,13 @@ REMOVE = 12
 CIRCLE = 0
 RECT = 1
 
-# pygame 1.9 <--> 2.0 compatibility
+# pygame 1.9 <--> 2.x compatibility
 SIZECHANGED = getattr(pygame, "WINDOWSIZECHANGED", None)
 WINEXPOSED = getattr(pygame, "WINDOWEXPOSED", None)
+EVENT_TYPE = pygame.event.EventType  # Could become deprecated?
 
-
-class PixelData:
-    "A class for storing, compressing, and converting raw pixel data"
-
-    def __str__(self):
-        name = type(self).__name__
-        kb = len(self._data) / 1024
-        return "<{0} {1} {3}x{4} [{2:.1f} kb]>".format(name, self.mode, kb, *self.size)
-
-    def __init__(self, img, compress=False, codec=zlib):
-        rgb = "RGBA", "RGB"
-        self.compressed = False
-        if type(img) is bytes: img = img[:-12], img[-12:]
-        elif type(img) is PixelData: img = img.raw()
-        elif isinstance(img, Graphic):
-            try: img = img.image
-            except: img = img.snapshot()
-        if type(img) is tuple:
-            m, w, h, c = self._unpack(img[1])
-            self.compressed = c
-            self._data = img[0]
-            self.size = w, h
-        elif isinstance(img, pygame.Surface):
-            self.size = img.get_size()
-            bits = img.get_bitsize()
-            m = "RGB" if bits == 24 else "RGBA" if bits == 32 else None
-            self._data = pygame.image.tostring(img, m)
-        else: # Pillow image
-            self.size = img.size
-            m = img.mode
-            if m not in rgb:
-                m = rgb[0]
-                img = img.convert(m)
-            self._data = img.tobytes()
-        if m in rgb: self.mode = m
-        else: raise NotImplementedError("Only RGB and RGBA modes are supported")
-        self.codec = codec
-        if compress: self.compress()
-        elif self.compressed: self.decompress()
-
-    def compress(self):
-        if not self.compressed:
-            self._data = self.codec.compress(self._data)
-            self.compressed = True
-        return self
-
-    def decompress(self):
-        if self.compressed:
-            self._data = self.codec.decompress(self._data)
-            self.compressed = False
-        return self
-
-    def _pack(self):
-        m = self.mode
-        m = 0 if m == "RGB" else 1
-        if self.compressed: m += 2
-        return struct.pack("!3I", m, *self.size)
-
-    @staticmethod
-    def _unpack(p):
-        m, w, h = struct.unpack("!3I", p)
-        c = bool(m & 2)
-        m = "RGBA" if (m & 1) else "RGB" 
-        return m, w, h, c
-
-    def raw(self): return self._data, self._pack()
-    
-    def __eq__(self, other): return self.raw() == other.raw()
-
-    def writeTo(self, f):
-        for b in self.raw(): f.write(b)
-        return self
-
-    def __bytes__(self):
-        b = self.raw()
-        return b[0] + b[1]
-
-    def _image(self, fn):
-        "Convert raw data to an image using the function provided"
-        data = self._data
-        if self.compressed: data = self.codec.decompress(data)
-        return fn(data, self.size, self.mode)
-
-    @property
-    def srf(self): return self._image(pygame.image.fromstring)
-
-    @property
-    def img(self): return Image(self.srf)
-
-    @staticmethod
-    def _frombytes(d, s, m):
-        return sys.modules["PIL.Image"].frombytes(m, s, d)
-
-    @property
-    def pil(self): return self._image(PixelData._frombytes)
+pygame.init()
+pygame.key.set_repeat(400, 80)
 
 
 class Graphic:
@@ -157,6 +65,7 @@ class Graphic:
     of the surface passed as an argument. Alternatively, the subclass may
     provide a 'image' property which gives a surface that Graphic.draw can use."""
     autoPositionOnResize = True
+    removeFrame = None
     _avgColor = None
     _preserve = "xy", "size"
     scrollable = True
@@ -166,10 +75,12 @@ class Graphic:
     angle = 0
     hoverable = True
     focusable = False
-    ondraw = None
     effects = None
     radiusFactor = 0.25
 
+    def update(self, ev):
+        if hasattr(self, "ondraw"): self.ondraw(ev)
+
     @property
     def name(self):
         "The key used when adding the instance to a canvas"
@@ -235,8 +146,8 @@ class Graphic:
 
     @property
     def aspectRatio(self):
-        size = self.size
-        return size[0] / size[1]
+        w, h = self.size
+        return w / h
 
     @size.setter
     def size(self, size): self.resize(size)
@@ -290,7 +201,7 @@ class Graphic:
         x += offset[0]
         y += offset[1]
         a = self.anchor
-        if a: # (blitSize[*] - 1) makes robot jiggly
+        if a:
             x -= blitSize[0] * (a & 3) // 2
             y -= blitSize[1] * (a & 12) // 8
         return x, y
@@ -298,24 +209,21 @@ class Graphic:
     def calcBlitRect(self, blitSize):
         cv = self.canvas
         offset = cv.rect.topleft if cv else (0,0)
-        return pygame.Rect(self.blitPosition(offset, blitSize), blitSize) ## !!!
+        return pygame.Rect(self.blitPosition(offset, blitSize), blitSize)
 
-    def relXY(self, pos):
+    def relPos(self, pos):
         "Calculate coordinates relative to the graphic object"
         if self.angle:
-            xc, yc = self.rect.center
-            x, y = transform2d(pos, preShift=(-xc,-yc), rotate=-self.angle)
-            xc, yc = self.center
-            return round(x + xc), round(y + yc)
+            x, y = transform2d(pos, shift1=neg(self.rect.center), matrix=-self.angle, shift2=self.center)
+            return round(x), round(y)
         else:
-            r = self.rect.topleft
-            return pos[0] - r[0], pos[1] - r[1]
+            return delta(pos, self.rect.topleft)
 
     def contains(self, pos):
         "Check if the graphic contains the coordinates"
         if self.angle:
             r = pygame.Rect((0,0), self.size)
-            pos = self.relXY(pos)
+            pos = self.relPos(pos)
         else: r = self.rect
         return bool(r.collidepoint(pos))
 
@@ -327,7 +235,6 @@ class Graphic:
                 setattr(self, a, (x * fx, y * fy))
             except: pass
 
-
 # Canvas interaction
 
     def _setCanvas(self, cv, key):
@@ -336,6 +243,10 @@ class Graphic:
             if key in cv and cv[key] is not self:
                 raise KeyError("Key '{}' is already in use".format(key))
             self._name = key
+        if cv.sketch is not None:
+            r = self.removeFrame
+            if r is not None and r < cv.sketch.frameCount:
+                raise ValueError("Can't add graphic with expired removeFrame value")
         self.canvas = cv
         cv._items.append(self)
 
@@ -360,12 +271,12 @@ class Graphic:
         except: pass
         return self
 
-    def toggle(self, name=None):
-        cv = self.canvas
-        if cv is not None:
-            if self in cv: self.remove()
-            else: self.setCanvas(cv, name)
-        return self
+#     def toggle(self, name=None):
+#         cv = self.canvas
+#         if cv is not None:
+#             if self in cv: self.remove()
+#             else: self.setCanvas(cv, name)
+#         return self
 
     @property
     def layer(self):
@@ -430,9 +341,11 @@ class Graphic:
     @property
     def dialog(self): return self.path[-2]
 
-    def bubble(self, eventName, ev):
+    def bubble(self, eventName, ev={}):
         "Pass an event to a different handler"
-        self.sketch.evMgr.handle(self, eventName, ev)
+        if not isinstance(ev, EVENT_TYPE):
+            ev = pygame.event.Event(pygame.USEREVENT, **ev)
+        return self.sketch.evMgr.handle(self, eventName, ev)
 
     @property
     def timeFactor(self):
@@ -454,7 +367,14 @@ class Graphic:
         srf = self.image
         if self.effects:
             srf = srf.copy()
-            for e in self.effects: srf = e.transition(srf, self.sketch.frameCount)
+            f = self.sketch.frameCount
+            rmv = []
+            for e in self.effects:
+                if f > e._t_max and e.remove: rmv.append(e)
+#                 if f <= e._t_max and f >= e._t_min:
+                else: srf = e.transition(srf, f)
+            if rmv:
+                for e in rmv: self.effects.remove(e)
         return srf
 
     def snapshot(self, **kwargs):
@@ -475,8 +395,7 @@ class Graphic:
             try: r = self.rect
             except: r = pygame.Rect((0,0), srf.get_size())
         if r.collidepoint(pos):
-            x, y = r.topleft
-            x, y = transform2d(pos, shift=(-x,-y))
+            x, y = delta(pos, r.topleft) 
             try: return srf.get_at((round(x), round(y)))
             except: pass
 
@@ -493,7 +412,7 @@ class Renderable(Graphic):
     def refresh(self):
         if self.stale: self.image
 
-    def remove(self, deleteRect=False): # !!!
+    def remove(self, deleteRect=False):
         super().remove(deleteRect)
         self.stale = True
         return self
@@ -528,7 +447,7 @@ class Renderable(Graphic):
         return self._srf.get_size()
 
     def resize(self, size):
-        self._size = size # ???
+        self._size = size
         self.stale = True
 
     @size.setter
@@ -540,7 +459,6 @@ class BaseSprite(Graphic):
     # Edge behaviours
     wrap = REMOVE
     bounce = bounceType = 0
-    onbounce = None
 
     # Kinematics
     spin = 0
@@ -562,7 +480,7 @@ class BaseSprite(Graphic):
         if p: p = (rgba(p[0]),) + p[1:]
         self._pen = p
 
-    def penReset(self):
+    def penReset(self, ev=None):
         p = self._pen
         if p: self._pen = p[:2] + (None,)
 
@@ -570,36 +488,32 @@ class BaseSprite(Graphic):
         self._size = size
         self.penReset()
 
-    def onwrap(self, update): self.penReset()
-
-    def _bounce(self, cv):
+    def _simpleBounce(self, cv):
         "Bounce sprite from the edge of its canvas"
         vx, vy = self.vel
         w, h = cv.size
         b = self.bounce
-        update = 0
+        where = CENTER
         if self.bounceType == 0:
             x, y = delta(self.rect.center, cv.rect.topleft)
             r = self.radius
             if b & HORIZONTAL and (x < r and vx < 0 or x > w-r and vx > 0):
                 self.vel = -vx, vy
-                update += HORIZONTAL
+                where += -1 if vx < 0 else 1
             if b & VERTICAL and (y < r and vy < 0 or y > h-r and vy > 0):
                 self.vel = vx, -vy
-                update += VERTICAL
+                where += -4 if vy < 0 else 4
         else:
             r = self.rect
             if b & HORIZONTAL and (r.left < 0 and vx < 0 or r.right >= w and vx > 0):
                 self.vel = -vx, vy
-                update += HORIZONTAL
+                where += -1 if x < r else 1
             if b & VERTICAL and (r.top < 0 and vy < 0 or r.bottom >= h and vy > 0):
                 self.vel = vx, -vy
-                update += VERTICAL  
-        return update
+                where += -4 if vy < 0 else 4
+        return where
 
-    circleBounce = _bounce
-
-    def simpleWrap(self, cv):
+    def _simpleWrap(self, cv):
         "Wrap sprite when it leaves the canvas"
         r = self.rect
         x, y = self.pos
@@ -609,32 +523,33 @@ class BaseSprite(Graphic):
                 dx, dy = cv.rect.topleft
                 r.move_ip((-dx, -dy))
             w = self.wrap
+            where = CENTER
             if w is True: w = BOTH
             if w & 5: # HORIZONTAL | REMOVE_X
                 d = r.width + cv.width
-                wrapX = True
-                if r.right < 0 and vx <= 0: x += d
-                elif r.left >= cv.width and vx >= 0: x -= d
-                else: wrapX = False
-                if wrapX and (w & 4):
+                if r.right < 0 and vx <= 0:
+                    x += d
+                    where -= 1
+                elif r.left >= cv.width and vx >= 0:
+                    x -= d
+                    where += 1
+                if where != CENTER and (w & 4):
                     self.remove()
                     return
-            else: wrapX = False
             if w & 10: # VERTICAL | REMOVE_Y
-                wrapY = True
                 d = r.height + cv.height
-                if r.bottom < 0 and vy <= 0: y += d
-                elif r.top >= cv.height and vy >= 0: y -= d
-                else: wrapY = False
-                if wrapY and (w & 8):
+                if r.bottom < 0 and vy <= 0:
+                    y += d
+                    where -= 4
+                elif r.top >= cv.height and vy >= 0:
+                    y -= d
+                    where += 4
+                if (w & 8) and where not in (4, 5, 6):
                     self.remove()
                     return
-            else: wrapY = False
-            if wrapX or wrapY:
-                update = HORIZONTAL if wrapX else 0
-                if wrapY: update += VERTICAL
-                self.onwrap(update)
             self.pos = x, y
+            if where != CENTER:
+                self.bubble("onwrap", {"where":where})
 
     def kinematics(self):
         "Update motion based on spin, vel, acc, and drag properties"
@@ -646,7 +561,7 @@ class BaseSprite(Graphic):
             self.pos = x + (vx + dvx / 2) * t, y + (vy + dvy / 2) * t
             self.vel = vx + dvx, vy + dvy
         else: self.pos = x + vx * t, y + vy * t
-        self.angle = positiveAngle(self.angle + self.spin * t)
+        if self.spin: self.angle = (self.angle + self.spin * t) % 360
         d = self.drag 
         if d:
             if type(d) in (int, float): s = d
@@ -655,19 +570,24 @@ class BaseSprite(Graphic):
             self.vel = d * vx, d * vy
             self.spin *= 1 - s 
 
-    def ondraw(self):
+    def motion(self):
         "Update sprite properties after drawing each frame"
         cv = self.canvas
         if self.bounce:
-            update = self._bounce(cv)
-            if update and self.onbounce: self.onbounce(update)
-        if self.wrap and self.simpleWrap(cv): return True
+            where = self._simpleBounce(cv)
+            if where != CENTER:
+                self.bubble("onbounce", {"where":where})
+        if self.wrap and self._simpleWrap(cv): return True
         self.kinematics()
         if self._pen:
             c, w, pos = self._pen
             if pos: cv._paint(pos, self.pos, c, w)
             self._pen = c, w, self.pos
 
+    def update(self, ev):
+        self.motion()
+        if hasattr(self, "ondraw"): self.ondraw(ev)
+
     def toward(self, pos, mag=None):
         "Return a vector directed toward the specified position"
         if mag is None: mag = hypot(*self.vel)
@@ -678,6 +598,10 @@ class Image(Graphic):
     "A class representing scaled and rotated images"
 
     def __init__(self, data=(2,2), bg=None):
+        try: # Convert bytes to Image
+            if type(data) is bytes or type(data[0]) is bytes:
+                data = Image.frombytes(data, False)
+        except: pass
         self._srf = CachedSurface(data, bg)
         self._size = self._srf.get_size()
 
@@ -690,10 +614,51 @@ class Image(Graphic):
             self._avgColor = pygame.transform.average_color(self._srf.original)
         return self._avgColor
 
-    def dumpCache(self): self._srf.dumpCache()
+    def dumpCache(self):
+        self._srf.dumpCache()
+        return self
+
+    def tobytes(self, info=False):
+        "Return the image as (bytes, ((w, h), mode)) where mode is 'RGB' or 'RGBA'"
+        srf = self.image
+        b = srf.get_bitsize()
+        if b not in (24, 32): raise ValueError("bit size must be 24 or 32")
+        mode = "RGB" if b == 24 else "RGBA"
+        data = pygame.image.tostring(srf, mode)
+        return (data, (srf.get_size(), mode)) if info else data
+
+    def __bytes__(self):
+        "Append the image mode and size info as a 12-byte struct to the raw image data"
+        data, info = self.tobytes(True)
+        mode = ("RGB", "RGBA").index(info[1])
+        return data + struct.pack("!3I", mode, *info[0])
 
     @staticmethod
-    def fromBytes(data): return PixelData(data).img
+    def frombytes(data, img=True):
+        if type(data) is bytes: data, info = data[:-12], data[-12:]
+        else: data, info = data
+        if type(info) is bytes:
+            mode, w, h = struct.unpack("!3I", info)
+            info = (w, h), "RGB" + ("", "A")[mode & 1]
+        srf = pygame.image.fromstring(data, *info)
+        return Image(srf) if img else srf
+
+    @staticmethod
+    def fromZip(key, archive=None):
+        if archive is None: archive = resolvePath("sc8pr.data")
+        return Image.frombytes(sc8prData(key, archive=archive))
+
+    def convert(self, alpha=False):
+        "Convert images to specified bit size"
+        img = self.image
+        srf = surface(img, alpha)
+        return self if srf is img else Image(srf)
+
+    @property
+    def rgb(self): return self.convert(False)
+
+    @property
+    def rgba(self): return self.convert(True)
 
     def tiles(self, cols=1, rows=1, flip=0, padding=0):
         "Create a list of images from a spritesheet"
@@ -727,9 +692,17 @@ class Image(Graphic):
         "Create a new image by flipping an existing instance"
         return Image(_pyflip(self.image, mode & HORIZONTAL, mode & VERTICAL))
 
-    def crop(self, *args):
+    def crop(self, *args, bg=True):
         "Create a new Image instance by cropping an existing image"
-        return Image(self.image.subsurface(pygame.Rect(*args)))
+        srf = self._srf.original
+        return Image(srf.subsurface(pygame.Rect(*args)) if args else crop(srf, bg)) 
+
+    def replace(self, oldColor=True, newColor=(255,255,255,0), dist=0.02):
+        "Copy the image and replace a (range of) colors"
+        srf = self._srf.original.copy()
+        if oldColor is True: oldColor = srf.get_at((0, 0))
+        pygame.PixelArray(srf).replace(oldColor, rgba(newColor), dist)
+        return Image(srf)
 
     @property
     def image(self):
@@ -745,18 +718,12 @@ class Image(Graphic):
         pygame.image.save(self._srf.original, fn)
         return self
 
-    def _export(self, fn="a.png"):
-        b = io.BytesIO(b"")
-        pygame.image.save(self.image, b, fn)
-        b.seek(0)
-        return b
+    @property
+    def png(self): return export(self.image).read()
 
     @property
-    def png(self): return self._export()
+    def jpg(self): return export(self.image, "a.jpg").read()
 
-    @property
-    def jpg(self): return self._export("a.jpg")
-        
     def copy(self): return Image(self.image.copy())
 
 
@@ -1004,7 +971,6 @@ class Canvas(Graphic):
         if mode & 2:
             br = isSketch and self.dirtyRegions is not None
             if br: self.dirtyRegions = []
-            ondrawList = self.sketch.ondrawList
             for g in list(self):
                 srf.set_clip(self.clipRect)
                 if not hasattr(g, "image") and g.effects:
@@ -1015,7 +981,6 @@ class Canvas(Graphic):
                 else: grect = g.draw(srf)
                 g.rect = grect
                 if br: self.dirtyRegions.append(grect)
-                if g.ondraw: ondrawList.append(g) # g.ondraw()
 
         # Draw border
         if mode & 1 and self.weight:
@@ -1077,7 +1042,7 @@ class Canvas(Graphic):
         for gr in self:
             yield gr
             if isinstance(gr, Canvas):
-                for i in gr.everything(): yield i
+                yield from gr.everything()
 
     def find(self, criteria, recursive=False):
         "Yield all Graphics that meet the criteria"
@@ -1110,13 +1075,12 @@ class Canvas(Graphic):
 
 class Sketch(Canvas):
     minSize = 32
-    capture = None
     realTime = False
     frameRate = 60
     _fixedAspect = True
     dirtyRegions = []
     resizeTrigger = False
-#     beforeResize = None
+    _sys_cursor = pygame.mouse.get_cursor()
 
     def __init__(self, size=(512,288)):
         super().__init__(size, "white")
@@ -1160,7 +1124,8 @@ class Sketch(Canvas):
         if self._fixedAspect and hasattr(bg, "aspectRatio"):
             self._fixedAspect = bg.aspectRatio
             size = self.size
-            self.resize(self._aspectSize(size, size))
+            size = self._aspectSize(size, size)
+            if size != self._size: self.resize(size)
         if self.dirtyRegions is not None:
             self.dirtyRegions = [pygame.Rect((0,0), self._size)]
 
@@ -1169,7 +1134,7 @@ class Sketch(Canvas):
 
     @cursor.setter
     def cursor(self, c):
-        if c is True: c = pygame.cursors.arrow
+        if c is True: c = Sketch._sys_cursor
         elif c is False: c = (8,8), (5,4), 8*(0,), 8*(0,)
         pygame.mouse.set_cursor(*c)
 
@@ -1212,7 +1177,7 @@ class Sketch(Canvas):
 
     def _pygameMode(self, n): return pygame.RESIZABLE if n is True else int(n)
 
-    def resize(self, size, mode=None): # Rewritten (non-recursive!) for v2.2.a3 !!!
+    def resize(self, size, mode=None):
         "Resize the sketch, maintaining aspect ratio if required"
         initSize = self._size
         ms = self.minSize
@@ -1222,12 +1187,12 @@ class Sketch(Canvas):
             size = round(f * size[0]), round(f * size[1])
         else:
             size = max(ms, round(size[0])), max(ms, round(size[1]))
+        if size == initSize: return # Prevent unnecessay resize!
         if mode is None: mode = self._mode
         else:
             mode = self._pygameMode(mode)
             self._mode = mode
         self.image = _pd.set_mode(size, mode)
-#         if self.beforeResize: self.beforeResize(initSize)
         super().resize(self.size)
         self._size = self.size
         if self.dirtyRegions is not None:
@@ -1245,13 +1210,12 @@ class Sketch(Canvas):
         "Initialize pygame and run the main drawing / event handling loop"
 
         # Initialize
-        pygame.init()
+        if not pygame.get_init(): pygame.init()
         self._clock = pygame.time.Clock()
-        pygame.key.set_repeat(400, 80)
         _pd.set_caption(caption)
         try:
-            try: icon = Image(icon) # pygame.image.load
-            except: icon = Image.fromBytes(sc8prData("alien"))
+            try: icon = Image(icon)
+            except: icon = Image.fromZip("alien")
             _pd.set_icon(icon.config(width=64).image)
         except: logError()
         w, h = self._size
@@ -1259,15 +1223,13 @@ class Sketch(Canvas):
         mode = self._pygameMode(mode)
         self._mode = mode
         self.image = _pd.set_mode(self._size, mode)
+        if self.image.get_bitsize() not in (24, 32):
+            print("Warning: Display may not be RGB/RGBA", file=sys.stderr)
         self.key = None
-        self.mouse = pygame.event.Event(pygame.USEREVENT,
-            code=None, pos=(0,0), description="Sketch startup")
+        self.mouse = customEv(code=None, pos=(0,0), description="Sketch startup")
 
         # Run setup
         try:
-            if hasattr(self, "_defer_coords"):
-                self.setCoords(*self._defer_coords)
-                del self._defer_coords
             if hasattr(self, "setup"): self.setup()
             else:
                 main = sys.modules["__main__"]
@@ -1280,19 +1242,21 @@ class Sketch(Canvas):
                 self.frameCount += 1
                 br = self.dirtyRegions
                 flip = br is None
-                self.ondrawList = []
                 self.draw()
                 if not flip:
                     br += self.dirtyRegions
                     flip = self._largeArea()
-                self._clock.tick(self.frameRate)
+
+                # _clock.tick was here in v2... better to update display first?
                 if flip: _pd.flip()
                 else: _pd.update(br)
-                self._capture()
-                for gr in self.ondrawList:
-                    try: gr.ondraw()
-                    except: logError()
-                if self.ondraw: self.ondraw()
+                self._clock.tick(self.frameRate)
+
+                for gr in list(self.everything()):
+                    gr.update(customEv(target=gr, handler="ondraw"))
+                    r = gr.removeFrame
+                    if r and self.frameCount >= r: gr.remove()
+                self.update(customEv(target=self, handler="ondraw"))
                 self._evHandle()
             except: logError()
 
@@ -1301,14 +1265,6 @@ class Sketch(Canvas):
         if mod: mod.Font.dumpCache()
         return self
 
-    def _capture(self):
-        "Call screen capture method when recording"
-        c = self.capture
-        if c is not None:
-            i = getattr(c, "interval", 1)
-            if self.frameCount % i == 0:
-                c.capture(self)
-
     def _evHandle(self):
         "Handle events in the pygame event queue"
         resized = False
