@@ -1,4 +1,4 @@
-# Copyright 2015-2020 D.G. MacCarthy <https://dmaccarthy.github.io/sc8pr>
+# Copyright 2015-2023 D.G. MacCarthy <https://dmaccarthy.github.io/sc8pr>
 #
 # This file is part of "sc8pr".
 #
@@ -52,6 +52,16 @@ class TextInput(Text):
         self.cursorStatus = False
         self.prompt = prompt
 
+    reformat = str
+
+    def _pw(self, t): return len(t) * self._pw_char
+
+    def hide(self, char="*"):
+        if char: self._pw_char = char[0]
+        try: self.bind(reformat=(self._pw if char else None))
+        except: pass
+        return self.config(stale=True)
+
     @property
     def highlight(self):
         if self._highlight: return self._highlight
@@ -93,7 +103,7 @@ class TextInput(Text):
             text = self.prompt
         else:
             color = self.color
-            text = self.data
+            text = self.reformat(self.data)
         try: srf = font.render(text, True, color)
         except:
             text = "[Unable to render!]"
@@ -236,14 +246,14 @@ class TextInput(Text):
 
     def _widthTo(self, i):
         font = Font.get(self.font, self.fontSize, self.fontStyle)
-        d = self.data
+        d = self.reformat(self.data)
         return (font.size(d[:i])[0] + font.size(d[:i+1])[0]) // 2
 
-    def onmousedown(self, ev): # onclick in v2.1
+    def onmousedown(self, ev):
         drag = ev.handler == "ondrag" 
         if drag or ev.button in self.allowButton:
             self._startCursor()
-            x = self.relXY(ev.pos)[0] - self.padding
+            x = self.relPos(ev.pos)[0] - self.padding
             n = len(self.data)
             i = 0
             while i < n and x > self._widthTo(i): i += 1
@@ -288,7 +298,7 @@ class TextInput(Text):
             else: pix = 0
         return pix, True
 
-    def scroll(self, pix=None, rel=True):
+    def _scroll(self, pix=None, rel=True):
         # Calculate scroll when not specified
         a = self.angle
         if pix is None: pix = self.focussed and a in (0, 90)
@@ -307,7 +317,9 @@ class TextInput(Text):
                 else: self.pos = sigma(self.pos, vec2d(pix, a))
         return self
 
-    ondraw = scroll
+    def update(self, ev):
+        self._scroll()
+        if hasattr(self, "ondraw"): self.ondraw(ev)
 
 
 class TextInputCanvas(Canvas):
