@@ -1,4 +1,4 @@
-# Copyright 2015-2021 D.G. MacCarthy <http://dmaccarthy.github.io>
+# Copyright 2015-2023 D.G. MacCarthy <http://dmaccarthy.github.io>
 #
 # This file is part of "sc8pr".
 #
@@ -31,12 +31,12 @@ class Font:
     _sort = None
     _cache = {}
     _cacheOrder = []
-    _serif = ("Merriweather", "DroidSerif", "Deja Vu Serif", "Palatino",
+    _serif = ("Merriweather", "DroidSerif", "DejaVuSerif", "Palatino",
         "Garamond", "Georgia", "Century", "TimesNewRoman", "Times")
-    _sans = ("LucidaSans", "OpenSans", "Oxygen", "Arsenal", "DroidSans", "Deja Vu Sans",
+    _sans = ("LucidaSans", "OpenSans", "Oxygen", "Arsenal", "DroidSans", "DejaVuSans",
         "MerriweatherSans", "Verdana", "Geneva", "Helvetica", "Arial")
-    _mono = ("SourceCodePro", "Inconsolata", "LucidaConsole",
-        "DroidSansMono", "Deja Vu Sans Mono", "Monaco", "CourierNew", "Courier")
+    _mono = ("Consolas", "Inconsolata", "LucidaConsole", "DroidSansMono",
+        "Deja Vu Sans Mono", "Monaco", "SourceCodePro", "CourierNew", "Courier")
 
     @staticmethod
     def _key(name, size=24, style=0):
