@@ -1,4 +1,4 @@
-# Copyright 2015-2022 D.G. MacCarthy <http://dmaccarthy.github.io>
+# Copyright 2015-2023 D.G. MacCarthy <http://dmaccarthy.github.io>
 #
 # This file is part of "sc8pr".
 #
@@ -16,7 +16,7 @@
 # along with "sc8pr".  If not, see <http://www.gnu.org/licenses/>.
 
 
-import sys, pygame
+import io, sys, pygame
 from random import randint
 from traceback import format_exc
 from zipfile import ZipFile
@@ -26,7 +26,10 @@ from sc8pr.geom import sigma, delta
 scale = pygame.transform.smoothscale
 
 def nothing(*args): pass
-  
+
+def customEv(**kwargs):
+    return pygame.event.Event(pygame.USEREVENT, **kwargs)
+
 def modKeys():
     key = pygame.key.get_pressed()
     m = 0
@@ -53,7 +56,6 @@ def _rgba(*args):
         t = type(c)
         if t is bool: # Random color
             c = [randint(0,255) for i in range(4 if c else 3)]
-##        yield pygame.Color(c) if t is str else pygame.Color(*c)
         yield _pgc(c)
 
 def rgba(*args):
@@ -115,6 +117,26 @@ def setAlpha(srf, a):
     srf.fill((255,255,255,a), special_flags=pygame.BLEND_RGBA_MIN)
     return srf
 
+_srf_rgba = pygame.Surface((2, 2), pygame.SRCALPHA, 32)
+_srf_rgb = pygame.Surface((2, 2), depth=24)
+_srf_val_err = ValueError("object cannot be converted to a Surface")
+
+def surface(srf, alpha=None):
+    "Convert something else to a Surface"
+    if not isinstance(srf, pygame.Surface):
+        if hasattr(srf, "image"): srf = srf.image
+        elif hasattr(srf, "snapshot"): srf = srf.snapshot().original
+        elif type(srf).__module__ == "PIL.Image":
+            try: srf = pygame.image.fromstring(srf.tobytes(), srf.size, srf.mode)
+            except: raise _srf_val_err
+        else:
+            try: srf = Image.frombytes(srf, False)
+            except: raise _srf_val_err
+    a = hasAlpha(srf), srf.get_bitsize()
+    if alpha is True and a != (True, 32): srf = srf.convert_alpha(_srf_rgba)
+    elif alpha is False and a != (False, 24): srf = srf.convert(_srf_rgb)
+    return srf
+
 def style(srf, bg=None, border=(0,0,0), weight=0, padding=0, borderradius=None):
     "Create a new surface with padding, background color, and/or border"
     w, h = srf.get_size()
@@ -194,6 +216,13 @@ def dragDrop(gr, ev):
     gr.config(pos=pos, hoverable=True)
     if drop: gr.bubble("ondrop", ev)
 
+def export(srf, fn="a.png"):
+    "Write a surface to a different format"
+    b = io.BytesIO(b"")
+    pygame.image.save(srf, b, fn)
+    b.seek(0)
+    return b
+
 def fileExt(fn, ext):
     "Force file extension"
     ext = [e.lower() for e in ([ext] if type(ext) is str else ext)]
@@ -213,6 +242,37 @@ def mix(x, y):
         for p in mix([x], y): yield p
 
 
+def _allPixels(px, c):
+    "Check if all pixels in a column are equal to the specified value"
+    return max(px) == c and min(px) == c
+
+def _h_crop(pxa, c):
+    "Find blank PixelArray columns"
+    n = len(pxa)
+    i = 0
+    j = n - 1
+    while i < n and _allPixels(pxa[i], c): i += 1
+    while j > i and _allPixels(pxa[j], c): j -= 1
+    return i, j, n
+
+def crop(srf, bg=True):
+    "Find a subsurface that crops the image to its content"
+    pxa = pygame.PixelArray(srf)
+    if bg is True: bg = pxa[0][0]
+    elif type(bg) is not int: bg = srf.map_rgb(pygame.Color(bg))
+    i, j, n = _h_crop(pxa, bg)
+    if i or j < n - 1:
+        h = srf.get_size()[1]
+        srf = srf.subsurface([i, 0, j - i + 1, h])
+        pxa = pygame.PixelArray(srf)
+    pxa = pxa.transpose()
+    i, j, n = _h_crop(pxa, bg)
+    if i or j < n - 1:
+        h = srf.get_size()[0]
+        srf = srf.subsurface([0, i, h, j - i + 1])
+    return srf
+
+
 class CachedSurface:
     "A class for caching scaled and rotated surfaces"
 
@@ -222,8 +282,8 @@ class CachedSurface:
         if t is str:
             srf = pygame.image.load(srf)
             if bg: srf = style(srf, bg)
-            elif srf.get_bitsize() < 32:
-                srf = srf.convert_alpha()
+# !!!            elif srf.get_bitsize() < 32: srf = srf.convert_alpha()
+            else: srf = surface(srf, True)
         elif t in (list, tuple):
             srf = pygame.Surface(srf, pygame.SRCALPHA)
             if bg is not None:
